import time
from queue import Queue
from . import progress_bar
from .PeerSeeker import PeerSeeker
from .PeerManager import PeerManager
from .Tracker import Tracker
from .WebSeedManager import WebSeedManager


class Client(object):
    @classmethod
    def __init__(self, torrent, downloaded_amount, piece_manager):
        self.torrent = torrent
        self.downloaded_amount = downloaded_amount
        self.start_time = time.time()
        self.new_peers_queue = Queue()
        self.request_queue = Queue()

        self.piece_manager = piece_manager
        self.peer_manager = PeerManager(torrent.urls, self.piece_manager, self.request_queue)
        self.peer_seeker = PeerSeeker(self.new_peers_queue, self.torrent, self.peer_manager)
        self.tracker = Tracker(torrent, self.new_peers_queue, downloaded_amount)

        self.tracker.start()
        self.peer_manager.start()
        self.peer_seeker.start()

        self.web_seed_managers = []
        num_web_seed_managers = len(self.peer_manager.http_peers) * 5
        for i in range(num_web_seed_managers):
            t = WebSeedManager(self.request_queue, self.peer_manager.http_peers)
            t.setDaemon(True)
            t.start()
            self.web_seed_managers.append(t)

    def start(self):
        while not self.piece_manager.finished():
            self.piece_manager.reset_pending()
            pieces_by_file = self.piece_manager.pieces_by_file()
            self.peer_manager.make_requests(pieces_by_file)
            self.peer_manager.enqueue_http_requests(pieces_by_file)

            # Record progress
            cur_downloaded = self.piece_manager.check_finished_pieces()
            rate = (cur_downloaded - self.downloaded_amount)/(time.time()-self.start_time)/1000. # rate in KBps
            progress_bar.print_progress(cur_downloaded, self.torrent.total_length, "BT:{}, Web:{}".format(len(self.peer_manager.peers), len(self.peer_manager.http_peers)), "({0:.2f}kB/s)".format(rate)) # + " Downloaded " + str(round(cur_downloaded/1000000., 2)) + "MB" )
            self.tracker.set_downloaded(cur_downloaded)
            time.sleep(0.1)

        cur_downloaded = self.piece_manager.check_finished_pieces()
        progress_bar.print_progress(cur_downloaded, self.torrent.total_length, "BT:{}, Web:{}".format(len(self.peer_manager.peers), len(self.peer_manager.http_peers)), "({0:.2f}kB/s)".format(rate)) # + " Downloaded " + str(round(cur_downloaded/1000000., 2)) + "MB" )

        print("\n Download Complete!") # . Downloaded " + str(cur_downloaded/1000000.) + " MB in " + str(time.time()-self.start_time) + " seconds.")
        self.piece_manager.close()
        self.tracker.request_stop()
        self.peer_seeker.request_stop()
        self.peer_manager.request_stop()
        for web_seed_manager in self.web_seed_managers:
            web_seed_manager.request_stop()



import math
import time
from . import utils
import itertools
from collections import defaultdict, OrderedDict
from pubsub import pub
from . import Block
BLOCK_SIZE = 2 ** 14


class Piece(object):
    def __init__(self, index, size, data_hash):
        self.index = index
        self.size = size
        self.data_hash = data_hash
        self.files_pending = {}
        self.files_finished = {}
        self.files = []
        self.BLOCK_SIZE = BLOCK_SIZE
        self.blocks = []
        self.init_blocks()

    def init_blocks(self):
        num_full_blocks = int(math.floor(float(self.size) / self.BLOCK_SIZE))
        self.blocks = []
        for _ in range(num_full_blocks):
            self.blocks.append(Block.Block(size=self.BLOCK_SIZE))
        if (self.size % BLOCK_SIZE) > 0:
            self.blocks.append(Block.Block(size=self.size % BLOCK_SIZE))

    def get_block_statuses(self):
        return [block.status for block in self.blocks]

    def set_file(self, filename, data):
        try:
            index = int(self.get_offset(filename) / BLOCK_SIZE)
            offset = int(self.get_offset(filename) % self.BLOCK_SIZE)
        except Exception:
            return
        done = 0
        while done != len(data):
            if offset != 0:
                self.blocks[index].status = "Partial"
                amount = min(self.BLOCK_SIZE - offset, len(data))
            elif len(data) - done < self.blocks[index].size:
                self.blocks[index].status = "Partial"
                amount = len(data) - done
            else:
                self.blocks[index].status = "Full"
                amount = self.blocks[index].size if len(data) - done > self.BLOCK_SIZE else len(data) - done
            self.blocks[index].data[offset] = data[done: done + amount]

            # increment counts
            done += amount
            offset = (offset + amount) % self.BLOCK_SIZE
            index += 1
        self.try_complete()

    def try_complete(self):
        if all([block.status == "Full" or block.status == "Partial" for block in self.blocks]):
            buf = bytearray(b"")
            for index in range(len(self.blocks)):
                buf.extend(self.blocks[index].assemble_data())
            if len(buf) > self.size:
                print("resetting,...")
                self.files_pending = {}
                # self.reset_pending_blocks()
                self.init_blocks()

            if utils.sha1_hash(buf) == self.data_hash:
                for filename in self.files_pending:
                    self.add_file_finished(filename)
                self.files_pending = {}
                self.writeFilesOnDisk(buf)
                pub.sendMessage('PieceManager.update_bitfield', index=self.index)

    def set_block(self, offset, data):
        index = int(offset / BLOCK_SIZE)
        self.blocks[index].data[offset % BLOCK_SIZE] = bytearray(data)
        self.blocks[index].status = "Full"
        self.try_complete()

    def set_all_blocks_pending(self):
        for block in self.blocks:
            block.set_pending()

    def reset_pending_blocks(self):
        for block in self.blocks:
            block.reset_pending()

    def set_file_pending(self, filename):
        self.files_pending[filename] = time.time()

    def remove_file_pending(self, filename):
        try:
            del self.files_pending[filename]
        except KeyError:
            pass

    def add_file_finished(self, filename):
        self.files_finished[filename] = True

    def isCompleteOnDisk(self):
        block_offset = 0
        data = b''
        for f in self.files:
            try:
                f_ptr = open(f["path"], 'rb')
            except IOError:
                break
            f_ptr.seek(f["file_offset"])
            data += f_ptr.read(f["length"])
            f_ptr.close()
            block_offset += f['length']

        if data and utils.sha1_hash(data) == self.data_hash:
            data = bytearray(b'')
            for block in self.blocks:
                block.status = "Full"
            return True
        return False

    def writeFunction(self, pathFile, data, offset):
        try:
            f = open(pathFile, 'r+b')
        except IOError:
            f = open(pathFile, 'wb')
        f.seek(offset)
        f.write(data)
        f.close()

    def writeFilesOnDisk(self, data):
        for f in self.files:
            pathFile = f["path"]
            file_offset = f["file_offset"]
            piece_offset = f["piece_offset"]
            length = f["length"]
            self.writeFunction(pathFile, data[piece_offset: piece_offset + length], file_offset)
        for block in self.blocks:
            block.data = {0: bytearray(b'')}

    def get_file_offset(self, filename):
        for f in self.files:
            if f.get('path').split('/')[-1] == filename:
                return f.get('file_offset')

    def get_length(self, filename):
        for f in self.files:
            if f.get('path').split('/')[-1] == filename:
                return f.get('length')

    def get_offset(self, filename):
        for f in self.files:
            if f.get('path').split('/')[-1] == filename:
                return math.floor(f.get('piece_offset'))



import time
import bencode
import logging
import os
import requests
from . import utils
try:
    from urllib.parse import urlparse, urlencode
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError


class Torrent(object):
    def __init__(self, hash, datastore):
        self.hash = hash
        self.datastore = utils.get_datastore(datastore=datastore)
        contents = ""
        if not os.path.isdir(self.datastore):
            os.makedirs(self.datastore)
        try:
            contents = open("/tmp/" + hash + '.torrent', 'rb').read()
        except Exception:
            contents = self.get_from_file()
            if not contents:
                contents = self.get_from_url()
            if not contents:
                raise Exception("Could not find a torrent with this hash on the tracker or in the data directory:" + str(self.datastore))

        self.contents = bencode.decode(contents)
        self.total_length = 0
        self.piece_length = self.contents['info']['piece length']
        self.pieces = self.contents['info']['pieces']

        self.info_hash = utils.sha1_hash(bencode.encode(self.contents['info']))
        self.peer_id = self.generate_peer_id()
        self.trackers = self.get_trackers()
        self.urls = self.get_urls()
        self.filenames = []

        self.get_files()
        if self.total_length % self.piece_length == 0:
            self.number_of_pieces = self.total_length / self.piece_length
        else:
            self.number_of_pieces = int(self.total_length / self.piece_length) + 1

        logging.debug(self.trackers)
        logging.debug(self.filenames)

        assert(self.total_length > 0)
        assert(len(self.filenames) > 0)

        name = self.contents['info']['name']
        if "length" in self.contents['info']:
            size_mb = self.contents['info']['length']/1000./1000.
        else:
            total_length = 0
            for f in self.contents['info']['files']:
                total_length += f['length']
            size_mb = total_length/1000./1000.

        print("Torrent name: " + name + ", Size: {0:.2f}MB".format(size_mb))

    def get_urls(self):
        urls = []
        for url in self.contents.get('url-list'):
            if not url:
                continue
            resp = requests.head(url)
            if resp.headers.get('Accept-Ranges', False):
                urls.append('/'.join(url.split('/')[0:-1]) + '/')
                continue

            directory = self.contents.get('info', {}).get('name')
            filename = self.contents.get('info', {}).get('files', [{'path': ['']}])[0].get('path')[0]
            if url[-1] == '/':
                url = url + directory + '/'
            else:
                url = url + '/' + directory + '/'

            resp = requests.head(url + filename)
            if resp.headers.get('Accept-Ranges', False):
                urls.append(url)
        return urls

    def get_from_file(self):
        torrent_path = os.path.join("/tmp/", self.hash + '.torrent')
        try:
            return open(torrent_path, 'rb').read()
        except Exception:
            return None
        return

    def get_from_url(self):
        contents = None
        url = "http://academictorrents.com/download/" + self.hash
        torrent_path = os.path.join("/tmp/", self.hash + '.torrent')
        response = urlopen(url, timeout=5).read()
        open(torrent_path, 'wb').write(response)
        try:
            return open(torrent_path, 'rb').read()
        except Exception:
            return None

    def get_files(self):
        root = self.datastore + self.contents['info']['name']
        if 'files' in self.contents['info']:
            if not os.path.exists(root):
                os.mkdir(root, 0o766)

            for f in self.contents['info']['files']:
                path = os.path.join(root, *f["path"])

                if not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path))

                self.filenames.append({"path": path, "length": f["length"]})
                self.total_length += f["length"]
        else:
            self.filenames.append({"path": root, "length": self.contents['info']['length']})
            self.total_length = self.contents['info']['length']

    def get_trackers(self):
        if 'announce-list' in self.contents:
            return self.contents['announce-list']
        else:
            return [[self.contents['announce']]]

    def generate_peer_id(self):
        seed = str(time.time())
        return utils.sha1_hash(seed.encode())



import hashlib
import os
import json
import datetime


def convert_bytes_to_decimal(headerBytes):
    size = 0
    power = len(headerBytes) - 1
    for ch in headerBytes:
        if isinstance(ch, int):
            size += ch * 256 ** power
        else:
            size += int(ord(ch)) * 256 ** power
        power -= 1
    return size


def sha1_hash(string):
    """Return 20-byte sha1 hash of string."""
    return hashlib.sha1(string).digest()


def get_timestamp_filename():
    return "/tmp/torrent_timestamps.json"


def get_datastore(datastore="", path_to_config_file="~/.academictorrents.config"):
    if datastore:
        datastore = clean_path(datastore)
    else:
        datastore = json.loads(open(clean_path(path_to_config_file)).read()).get("datastore", os.getcwd() + "/datastore/")
    if datastore[-1] != "/":
        datastore = datastore + "/"
    return datastore


def clean_path(path=None):
    if path.startswith("~"):
        return os.path.expanduser(path)
    else:
        return os.path.abspath(path)

def write_timestamp(at_hash):
    filename = get_timestamp_filename()
    with open(filename, 'w+') as f:
        try:
            timestamps = json.loads(f.read())
        except ValueError:
            timestamps = {}
        timestamps[at_hash] = int(datetime.datetime.now().strftime("%s"))
        json.dump(timestamps, f)


def read_timestamp(at_hash):
    filename = get_timestamp_filename()
    try:
        with open(filename) as f:
            return json.load(f).get(at_hash, 0)
    except Exception:
        return 0


def timestamp_is_within_30_days(timestamp):
    seconds_in_a_month = 86400 * 30
    if timestamp > int(datetime.datetime.now().strftime("%s")) - seconds_in_a_month:
        return True
    return False


def timestamp_is_within_10_seconds(timestamp):
    ten_seconds = 10
    if timestamp > int(datetime.datetime.now().strftime("%s")) - ten_seconds:
        return True
    return False


def filenames_present(torrent):
    return torrent.contents['info']['name'] in os.listdir(torrent.datastore)



# -*- coding: utf-8 -*-
import sys


# Print iterations progress
def print_progress(iteration, total, prefix='', suffix='', decimals=1, bar_length=50):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """
    if total != 1:
        total = total-1
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '█' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix))
    
    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()



import bencode
import requests
import logging
import struct
import random
import socket
import threading
import datetime
import time
from threading import Thread
from . import utils
from .version import __version__
# Python 2 and 3: alternative 4
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


class FuncThread(threading.Thread):
    def __init__(self, target, *args):
        threading.Thread.__init__(self)
        self._target = target
        self._args = args

    def run(self):
        self._target(*self._args)


class Tracker(Thread):
    def __init__(self, torrent, new_peers_queue, downloaded):
        Thread.__init__(self)
        self.torrent = torrent
        self.lstThreads = []
        self.new_peers_queue = new_peers_queue
        self.stop_requested = False
        self.setDaemon(True)
        self.downloaded = downloaded
        self.last_message_time = int(datetime.datetime.now().strftime("%s"))
        self.last_update_time = int(datetime.datetime.now().strftime("%s"))

    def request_stop(self):
        self.stop_requested = True

    def run(self):
        while not self.stop_requested:
            self.getPeersFromTrackers()
            time.sleep(10)
            self.downloading_message()
        self.stop_message()

    def getPeersFromTrackers(self):
        if utils.timestamp_is_within_10_seconds(self.last_update_time):
            return
        self.last_update_time = int(datetime.datetime.now().strftime("%s"))

        for tracker in self.torrent.trackers:
            if tracker[0] == '':
                continue
            elif tracker[0][:4] == "http":
                t1 = FuncThread(self.scrapeHTTP, self.torrent, tracker[0])
                t1.setDaemon(True)
                self.lstThreads.append(t1)
                t1.start()
            else:
                t2 = FuncThread(self.scrape_udp, self.torrent, tracker[0])
                t2.setDaemon(True)
                self.lstThreads.append(t2)
                t2.start()

        for t in self.lstThreads:
            t.join()

    def request_stop(self):
        self.stop_requested = True

    def set_downloaded(self, size):
        self.downloaded = size

    def scrapeHTTP(self, torrent, tracker):
        params = {
            'info_hash': torrent.info_hash,
            'peer_id': torrent.peer_id,
            'uploaded': 0,
            'downloaded': 0,
            'left': torrent.total_length,
            'event': 'started',
            'port': 6881
        }
        try:
            answerTracker = requests.get(tracker, params=params, timeout=30, headers={'user-agent': "AT-Client/" + __version__ + " " + requests.utils.default_user_agent()})
            lstPeers = bencode.decode(answerTracker.content)
            for peer in lstPeers['peers']:
                self.new_peers_queue.put([peer['ip'], peer['port']])
        except Exception as e:
            logging.info(e)
            pass

    def stop_message(self):
        resp = requests.models.Response()
        for tracker in self.torrent.trackers:
            if tracker[0] == '':
                continue
            elif tracker[0][:4] == "http":
                event = "completed" if self.torrent.total_length == self.downloaded else "stopped"
                params = {
                    'info_hash': self.torrent.info_hash,
                    'peer_id': self.torrent.peer_id,
                    'uploaded': 0,
                    'downloaded': self.downloaded,
                    'left': 0,
                    'event': event,
                    'port': 6881
                }
                try:
                    resp = requests.post(tracker[0], params=params, timeout=30, headers={'user-agent': "AT-Client/" + __version__ + " " + requests.utils.default_user_agent()})
                except Exception as e:
                    pass
            return params, resp

    def downloading_message(self):
        if utils.timestamp_is_within_10_seconds(self.last_message_time):
            return
        self.last_message_time = int(datetime.datetime.now().strftime("%s"))
        resp = requests.models.Response()
        if self.downloaded == 0:
            return True
        for tracker in self.torrent.trackers:
            if tracker[0] == '':
                continue
            elif tracker[0][:4] == "http":
                params = {
                    'info_hash': self.torrent.info_hash,
                    'peer_id': self.torrent.peer_id,
                    'uploaded': 0,
                    'downloaded': self.downloaded,
                    'left': self.torrent.total_length - self.downloaded,
                    'port': 6881
                }
                try:
                    resp = requests.get(tracker[0], params=params, timeout=30, headers={'user-agent': "AT-Client/" + __version__ + " " + requests.utils.default_user_agent()})
                except Exception as e:
                    logging.info(e)
                    pass
            return params, resp

    def make_connection_id_request(self):
        conn_id = struct.pack('>Q', 0x41727101980)
        action = struct.pack('>I', 0)
        trans_id = struct.pack('>I', random.randint(0, 100000))
        return (conn_id + action + trans_id, trans_id, action)

    def make_announce_input(self, info_hash, conn_id, peer_id):
        action = struct.pack('>I', 1)
        trans_id = struct.pack('>I', random.randint(0, 100000))

        downloaded = struct.pack('>Q', 0)
        left = struct.pack('>Q', 0)
        uploaded = struct.pack('>Q', 0)

        event = struct.pack('>I', 0)
        ip = struct.pack('>I', 0)
        key = struct.pack('>I', 0)
        num_want = struct.pack('>i', -1)
        port = struct.pack('>h', 8000)

        msg = (conn_id + action + trans_id + info_hash + peer_id + downloaded +
               left + uploaded + event + ip + key + num_want + port)

        return msg, trans_id, action

    def send_msg(self, conn, sock, msg, trans_id, action, size):
        sock.sendto(msg, conn)
        try:
            response = sock.recv(2048)
        except socket.timeout as err:
            logging.info(err)
            return
            # logging.debug("Connecting again...")
            # return self.send_msg(conn, sock, msg, trans_id, action, size)
        if len(response) < size:
            logging.debug("Did not get full message. Connecting again...")
            return self.send_msg(conn, sock, msg, trans_id, action, size)

        if action != response[0:4] or trans_id != response[4:8]:
            logging.debug("Transaction or Action ID did not match. Trying again...")
            return self.send_msg(conn, sock, msg, trans_id, action, size)

        return response

    def scrape_udp(self, torrent, announce):
        try:
            parsed = urlparse(announce)
            ip = socket.gethostbyname(parsed.hostname)

            if ip == '127.0.0.1':
                return False
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.settimeout(30)
            conn = (ip, parsed.port)
            msg, trans_id, action = self.make_connection_id_request()
            response = self.send_msg(conn, sock, msg, trans_id, action, 16)
            if response is None:
                return ""

            conn_id = response[8:]
            msg, trans_id, action = self.make_announce_input(torrent.info_hash, conn_id, torrent.peer_id)
            response = self.send_msg(conn, sock, msg, trans_id, action, 20)
            if response is None or response is "":
                return ""
            peersByte = response[20:]
            raw_bytes = [ord(c) for c in peersByte]
            for i in range(len(raw_bytes) / 6):
                start = i * 6
                end = start + 6
                ip = ".".join(str(i) for i in raw_bytes[start:end - 2])
                port = raw_bytes[end - 2:end]
                port = port[1] + port[0] * 256
                self.new_peers_queue.put([ip, port])
        except Exception:
            pass



import requests
import math
from pubsub import pub
from threading import Thread
import urllib3
import logging
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class HttpPeer(object):
    def __init__(self, url):
        self.url = url
        self.sess = requests.Session()
        self.fail_files = []

    def request_ranges(self, filename, pieces):
        start = pieces[0].get_file_offset(filename)
        end = start
        for piece in pieces:
            end += piece.get_length(filename)
        try:
            return self.sess.get(self.url + filename, headers={'Range': 'bytes=' + str(start) + '-' + str(end)}, verify=False, timeout=5)
        except Exception as e:
            logging.info(str(e))
            return False

    def publish_responses(self, response, filename, pieces):
        offset = 0
        for piece in pieces:
            size = piece.get_length(filename)
            pub.sendMessage('PieceManager.receive_file', piece=(piece.index, filename, response.content[offset: offset + size]))
            offset += size



import logging
import json
from queue import Queue
from .PieceManager import PieceManager
from .Torrent import Torrent
from .Client import Client
from .utils import read_timestamp, timestamp_is_within_30_days, filenames_present, write_timestamp, clean_path


def get(at_hash, datastore="~/.academictorrents-datastore", urls=[], showlogs=False):
    logging.getLogger().setLevel(logging.CRITICAL)
    if showlogs:
        logging.getLogger().setLevel(level=logging.INFO)

    torrent = Torrent(at_hash, datastore)
    torrent.urls = torrent.urls + urls
    path = torrent.datastore + torrent.contents['info']['name']

    # Check timestamp
    timestamp = read_timestamp(at_hash)
    if timestamp_is_within_30_days(timestamp) and filenames_present(torrent):
        return path

    # Check if downloaded and finished
    piece_manager = PieceManager(torrent)
    piece_manager.check_disk_pieces()
    downloaded_amount = piece_manager.check_finished_pieces()
    if float(downloaded_amount) / torrent.total_length == 1.0:
        return path

    print("Downloading to " + path)
    Client(torrent, downloaded_amount, piece_manager).start()
    write_timestamp(at_hash)
    return path

def set_datastore(datastore, path_to_config_file="~/.academictorrents.config"):
    json.dump({"datastore": datastore}, open(clean_path(path_to_config_file), "w+"))



from threading import Thread
import urllib3
from . import HttpPeer
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class WebSeedManager(Thread):
    def __init__(self, request_queue, http_peers):
        Thread.__init__(self)
        self.request_queue = request_queue
        self.stop_requested = False
        self.http_peers = http_peers
        self.setDaemon(True)

    def run(self):
        while not self.stop_requested:
            httpPeer, filename, pieces = self.request_queue.get()
            if not pieces:
                self.request_queue.task_done()
                continue
            response = httpPeer.request_ranges(filename, pieces)
            if not response or response.status_code != 206:
                httpPeer.fail_files.append(filename)
                for http_peer in self.http_peers:
                    response = http_peer.request_ranges(filename, pieces)
                    if response and response.status_code == 206:
                        httpPeer.publish_responses(response, filename, pieces)
                        break
                self.request_queue.task_done()
                continue
            httpPeer.publish_responses(response, filename, pieces)
            self.request_queue.task_done()

    def request_stop(self):
        self.stop_requested = True



import select
import struct
from queue import Queue
try:
    from itertools import zip_longest as zip_longest
except:
    from itertools import izip_longest as zip_longest
import time
import logging
from threading import Thread
from pubsub import pub
from . import utils
from . import RarestPieces
from .HttpPeer import HttpPeer


class PeerManager(Thread):
    def __init__(self, urls, piece_manager, request_queue):
        Thread.__init__(self)
        self.peers = []
        self.http_peers = []
        self.request_queue = request_queue
        self.piece_manager = piece_manager
        self.rarestPieces = RarestPieces.RarestPieces(piece_manager)
        self.stop_requested = False
        self.setDaemon(True)

        self.pieces_by_peer = []
        for i in range(self.piece_manager.number_of_pieces):
            self.pieces_by_peer.append([0, []])
        for url in urls:
            self.http_peers.append(HttpPeer(url))

        # Events
        pub.subscribe(self.add_peer, 'PeerManager.new_peer')
        pub.subscribe(self.peers_bitfield, 'PeerManager.update_peers_bitfield')

    def run(self):
        while not self.stop_requested:
            self.start_connection_to_peers()
            read = [p.socket for p in self.peers]
            readList, _, _ = select.select(read, [], [], 1)

            # Receive from peers
            for socket in readList:
                peer = self.get_peer_by_socket(socket)
                try:
                    msg = socket.recv(1024)
                except Exception as e:
                    logging.info(peer.ip + ": removing peer because of: " + str(e))
                    self.remove_peer(peer)
                    continue

                if not msg:
                    logging.error(peer.ip + ": removing peer because we received a message of 0 length")
                    self.remove_peer(peer)
                    continue

                peer.buffer += msg

                if not peer.has_handshaked:
                    peer.check_handshake()
                    continue

                if len(peer.buffer) < 4 or struct.unpack("!I", peer.buffer[:4])[0] == 0:  # keep alive
                    continue

                msg_length = utils.convert_bytes_to_decimal(peer.buffer[0:4])
                while len(peer.buffer) >= 4 + msg_length:
                    msg_code = int(ord(peer.buffer[4:5]))
                    payload = peer.buffer[5:4 + msg_length]
                    peer.buffer = peer.buffer[4 + msg_length:]
                    peer.id_function[msg_code](payload)

    def request_stop(self):
        self.stop_requested = True

    def peers_bitfield(self, bitfield=None, peer=None, index=None):
        if index is not None:
            self.pieces_by_peer[index] = ["", []]
            return

        for i in range(len(self.pieces_by_peer)):
            if bitfield[i] == 1 and peer not in self.pieces_by_peer[i][1] and not self.pieces_by_peer[i][0] == "":
                self.pieces_by_peer[i][1].append(peer)
                self.pieces_by_peer[i][0] = len(self.pieces_by_peer[i][1])

    def get_unchoked_peer(self, index):
        for peer in self.peers:
            if peer.bitfield[index] and not peer.state["peer_choking"]:
                return peer
        return False

    def start_connection_to_peers(self):
        peers_to_handshake = [peer for peer in self.peers if not peer.has_handshaked and peer.last_handshake_attempt < time.time() - 10]
        for peer in peers_to_handshake:
            peer.last_handshake_attempt = time.time()
            try:
                peer.handshake()
                interested = peer.build_interested()
                peer.send(interested)
            except Exception as e:
                logging.error(peer.ip + ": removing peer because of: " + str(e))
                self.remove_peer(peer)

    def add_peer(self, peer):
        self.peers.append(peer)

    def remove_peer(self, peer):
        if peer in self.peers:
            try:
                peer.socket.close()
            except Exception:
                pass
            self.peers.remove(peer)

        for rarestPiece in self.rarestPieces.rarestPieces:
            if peer in rarestPiece["peers"]:
                rarestPiece["peers"].remove(peer)

    def get_peer_by_socket(self, socket):
        for peer in self.peers:
            if socket == peer.socket:
                return peer
        raise ("peer not present in PeerList")

    def request_new_piece(self, peer, index, offset, length):
        request = peer.build_request(index, offset, length)
        peer.send(request)

    def make_requests(self, pieces_by_file):
        max_requests = 50
        requests = 0
        i = 0
        if not self.peers:
            return
        pieces = [pieces for _, pieces in pieces_by_file for pieces in pieces]
        while i < len(pieces) and requests < max_requests:
            piece = pieces[i]
            peer = self.get_unchoked_peer(piece.index)
            if not peer:
                i += 1
                continue
            for idx, block in enumerate(piece.blocks):
                if block.status == "Free" or block.status == "Partial":
                    block.set_pending()
                    self.request_new_piece(peer, piece.index, idx * piece.BLOCK_SIZE, block.size)
                    requests += 1
            i += 1

    def enqueue_http_requests(self, pieces_by_file):
        i = 0
        if not self.request_queue.empty() or not self.http_peers:
            return
        while pieces_by_file:
            filename, pieces_containing_file = pieces_by_file.pop()
            for pieces in grouper(pieces_containing_file):
                self.piece_manager.set_pending(filename, pieces)
                for peer in self.http_peers:
                    if filename not in peer.fail_files:
                        self.request_queue.put((peer, filename, pieces), False)
                        i += 1
                        break

def grouper(pieces):
    resp = []
    length = 0
    temp_pieces = []
    for piece in pieces:
        temp_pieces.append(piece)
        if length > 5000000:
            resp.append(temp_pieces)
            temp_pieces = []
        length += piece.size
    resp.append(temp_pieces)

    # now we have a list of piece lists
    for idx, pieces in enumerate(resp):
        resp[idx] = [piece for piece in pieces if piece]
    resp = resp if resp[0] else []
    return resp



from . import Piece
import bitstring
import time
from threading import Thread
from pubsub import pub
from . import progress_bar
from collections import defaultdict
import math
import random


class PieceManager(object):
    def __init__(self, torrent):
        self.torrent = torrent
        self.number_of_pieces = torrent.number_of_pieces
        self.bitfield = bitstring.BitArray(self.number_of_pieces)
        self.pieces = self.generate_pieces()
        self.files = self.get_files()
        for f in self.files:
            id_piece = f['id_piece']
            self.pieces[id_piece].files.append(f)

        # Create events
        pub.subscribe(self.receive_block, 'PieceManager.receive_block')
        pub.subscribe(self.receive_file, 'PieceManager.receive_file')
        pub.subscribe(self.update_bitfield, 'PieceManager.update_bitfield')


    def close(self):
        pub.unsubAll("PieceManager.receive_block")
        pub.unsubAll("PieceManager.receive_file")
        pub.unsubAll("PieceManager.update_bitfield")
        self.torrent = None
        self.bitfield = None
        self.pieces = None
        self.files = None

    def finished(self):
        if sum(self.bitfield) == self.number_of_pieces:
            return True
        return False

    def check_finished_pieces(self):
        b = 0
        for i in range(self.number_of_pieces):
            if self.bitfield[i]:
                b += self.pieces[i].size
        return b

    def check_full_blocks(self):
        b = 0
        for piece in self.pieces:
            for block in piece.blocks:
                if block.status == "Full":
                    b += block.size
        return b

    def update_bitfield(self, index):
        self.bitfield[index] = 1

    def receive_block(self, piece):
        index, offset, data = piece
        self.pieces[index].set_block(offset, data)

    def receive_file(self, piece):
        index, filename, data = piece
        try:
            self.pieces[index].set_file(filename, data)
        except Exception:
            pass

    def reset_pending(self):
        for piece in self.pieces:
            for block in piece.blocks:
                block.reset_pending()

    def set_pending(self, filename, pieces):
        for piece in pieces:
            cur = int(math.floor(piece.get_offset(filename)/piece.BLOCK_SIZE))
            last = int(math.floor((piece.get_length(filename) + piece.get_offset(filename))/piece.BLOCK_SIZE))
            piece.set_file_pending(filename)
            while cur < last:
                piece.blocks[cur].set_pending()
                cur += 1

    def pieces_by_file(self, reverse=False):
        pieces_by_file = defaultdict(list)
        pieces = [piece for index, piece in enumerate(self.pieces) if not self.bitfield[index]]
        for piece in pieces:
            for f in piece.files:
                filename = f.get('path').split('/')[-1]
                timestamp = piece.files_pending.get(filename, 0)
                if int(time.time() - timestamp) > 1 and filename not in piece.files_finished.keys():
                    pieces_by_file[filename].append(piece)
        sorted_by_length = sorted(pieces_by_file.items(), key=lambda k_v: len(k_v[1]), reverse=reverse)
        return sorted_by_length

    def check_disk_pieces(self):
        i = 0
        while i < self.number_of_pieces:
            self.bitfield[i] = self.pieces[i].isCompleteOnDisk()  # this should set all the finished bools on the finished pieces
            i += 1
            if i % 50 == 0 or i == self.number_of_pieces:
                progress_bar.print_progress(i - 1, self.number_of_pieces, "Checking for pieces on disk:", "")
        print("Found " + str(sum(self.bitfield)) + " finished pieces out of " + str(len(self.bitfield)) + " total pieces.")

    def generate_pieces(self):
        pieces = []
        for i in range(self.number_of_pieces):
            start = i * 20
            end = start + 20

            if i == (self.number_of_pieces - 1):
                length = self.torrent.total_length - (self.number_of_pieces-1) * self.torrent.piece_length
                pieces.append(Piece.Piece(i, length, self.torrent.pieces[start:end]))
            else:
                pieces.append(Piece.Piece(i, self.torrent.piece_length, self.torrent.pieces[start:end]))
        return pieces

    def get_files(self):
        files = []
        piece_offset = 0
        size_used = 0

        for f in self.torrent.filenames:

            current_size_file = f["length"]
            file_offset = 0

            while current_size_file > 0:
                id_piece = int(piece_offset / self.torrent.piece_length)
                size = self.pieces[id_piece].size - size_used

                if current_size_file - size < 0:
                    file = {"length": current_size_file,"id_piece":id_piece ,"file_offset":file_offset, "piece_offset":size_used ,"path":f["path"]}
                    piece_offset += current_size_file
                    file_offset += current_size_file
                    size_used += current_size_file
                    current_size_file = 0

                else:
                    current_size_file -= size
                    file = {"length":size,"id_piece":id_piece ,"file_offset":file_offset,"piece_offset":size_used , "path":f["path"]}
                    piece_offset += size
                    file_offset += size
                    size_used = 0

                files.append(file)
        return files



import time
from collections import defaultdict

class Block(object):
    def __init__(self, size, time=time.time(), status="Free", data=None):
        self.size = size
        self.status = status
        self.pending = False
        self.data = defaultdict(list)
        self.time = time

    def assemble_data(self):
        buf = bytearray(self.size)
        length = 0
        for index, data in self.data.items():
            index = int(index % self.size)
            buf[index: index + len(data)] = data
            length += len(data)
        if length != self.size:
            return bytearray(b"")
        return buf

    def set_pending(self):
        self.pending = True
        self.time = int(time.time())

    def reset_pending(self):
        if(int(time.time()) - self.time) > 8:
            self.pending = False
            self.time = 0



import time
from . import Peer
import logging
from threading import Thread
from pubsub import pub


class PeerSeeker(Thread):
    def __init__(self, new_peers_queue, torrent, peer_manager):
        Thread.__init__(self)
        self.stop_requested = False
        self.peer_manager = peer_manager
        self.new_peers_queue = new_peers_queue
        self.torrent = torrent
        self.setDaemon(True)
        self.reset_time = time.time()

    def request_stop(self):
        self.stop_requested = True

    def run(self):
        failed_peers = []
        while not self.stop_requested:
            # reset failed peers so we can try again
            if (time.time() - self.reset_time) > 10:
                failed_peers = []
                self.reset_time = time.time()

            peer = self.new_peers_queue.get()
            peer = Peer.Peer(self.torrent, peer[0], peer[1])
            extant_peers = [(peer.ip, peer.port) for peer in self.peer_manager.peers]
            if (peer.ip, peer.port) in failed_peers or (peer.ip, peer.port) in extant_peers:
                continue

            if peer.connect(5):
                pub.sendMessage('PeerManager.new_peer', peer=peer)
            else:
                failed_peers.append((peer.ip, peer.port))
